@@ -1,6 +1,5 @@
-/* $Id: puf.h,v 1.6 2002/05/23 08:43:01 ossi Exp $ *
- *
- * puf 0.9  Copyright (C) 2000-2002 by Oswald Buddenhagen <puf@ossi.cjb.net>
+/* 
+ * Copyright (C) 2000-2004 by Oswald Buddenhagen <puf@ossi.cjb.net>
  * based on puf 0.1.x (C) 1999,2000 by Anders Gavare <gavare@hotmail.com>
  *
  * You may modify and distribute this code under the terms of the GPL.
@@ -17,7 +16,16 @@
 
 #ifdef USE_LL
 # define _FILE_OFFSET_BITS 64
+# define SSOFFT "lli"
+#else
+# define SSOFFT "li"
 #endif
+#define SOFFT "%"SSOFFT
+
+#ifndef _O_BINARY
+# define _O_BINARY 0
+#endif
+
 #include <sys/types.h>                             
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -48,9 +56,6 @@
 /*  Undef this, if you don't want the -d switch  */
 #define DEBUG
 
-/*  This may be defined if you want to, but they are not neccessary:  */
-/*#define	HTTP_ACCEPT	"Accept: * / *"  */
-
 /*  Undef this, if you want no message about incorrect DNS resolver configs  */
 #define CORRECT_DNS
 
@@ -64,6 +69,9 @@
 /*  Maximal allowed header length  */
 #define MAXHEADERLEN			0x10000
 
+/*  Max file size when using -xo/-xO or multi-src targets. In MB.  */
+#define DEFAULT_MAX_BUFFER		10
+
 /*  Filename of a url ending with a slash:  */
 #define	DEFAULT_INDEX_FILE_NAME		"index.html"
 
@@ -92,21 +100,10 @@
 /*  Default Max number of connection attempts per url:  */
 #define	DEFAULT_MAX_ATTEMPTS		5
 
-#ifdef USE_LL
-# define SSOFFT "lli"
-#else
-# define SSOFFT "li"
-#endif
-#define SOFFT "%"SSOFFT
-
 #ifndef HAVE_STRDUP
 char *strdup(const char *s);
 #endif
 
-#ifdef USE_DMALLOC
-# include <dmalloc.h>
-#endif
-
 #if 1
 # define RND(m) ((rand() >> 15) * m / ((RAND_MAX >> 15) + 1)) /* m < 32768 */
 #else
@@ -120,7 +117,7 @@ typedef struct haddr_t {
     time_t retry_time;
     u_int
 	attempt:8,
-	last_errt:2,
+	last_errt:3,
 	err_wait:16;
 } haddr_t;
 
@@ -138,27 +135,34 @@ typedef struct hinfo_t {
 typedef struct host_t {
     struct host_t *next;
 
-    hinfo_t *info;
+    hinfo_t *info;		/*  whost_t* if !ready  */
+    u_char ready:1;
     char name[0];
 } host_t;
 
 
 /*  proxy spec.  */
-
 typedef struct proxy_t {
     struct proxy_t *next;	/*  next proxy in chain  */
 
     host_t *host;		/*  Proxy host  */
-    u_int
-	have_auth:1,		/*  User:Password (BASE64) after cgi_path?  */
-	ready:1;		/*  DNS lookup complete?  */
     u_short port;		/*  Proxy port  */
-    u_short ratio;		/*  Requested load ratio  */
-    int score;			/*  Connection/byte count - for capacity-driven balancing  */
-/*    u_short cur_conn;*/		/*  Currently open connections - for load-driven balancing - now unused */
+    u_char have_auth:1;		/*  User:Password (BASE64) after cgi_path?  */
     char cgi_path[0];		/*  Only for "cgi-proxies"  */
 } proxy_t;
 
+typedef struct proxyent_t {
+    proxy_t *proxy;
+    int score;			/*  Connection/byte count - for capacity-driven balancing  */
+    u_short ratio;		/*  Requested load ratio  */
+/*    u_short cur_conn;*/		/*  Currently open connections - for load-driven balancing - now unused */
+} proxyent_t;
+
+typedef struct proxyarr_t {
+    u_int nents, rents, spare:31, cow:1;	/* BC with ptrarr_t */
+    proxyent_t ents[0];
+} proxyarr_t;
+
 
 /*  fake user agent  */
 typedef struct agent_t {
@@ -169,37 +173,46 @@ typedef struct agent_t {
 
 /*  generic array of pointers  */
 typedef struct ptrarr_t {
-    void *ents;
-    u_int nents, rents, cow:1;
+    u_int nents, rents, spare:31, cow:1;
+    void *ents[0];
 } ptrarr_t;
 
 
-/*  disposition path spec.  */
+/*  disposition path spec. for -P  */
 typedef struct disp_path_t {
-    int file_num;		/* for -xe */
+    int file_num;		/* for -xE */
     char path[0];
 } disp_path_t;
 
 
-/*  url spec.  */
+/*  extension/pattern/mimetype spec. for -A/-R  */
+typedef struct filter_t {
+    u_int acc:1;
+    u_int type:1;
+    u_int pat:1;
+    char data[0];
+} filter_t;
 
+/*  command line options  */
 typedef struct options_t {
-#ifdef USE_MAGIC
-    int chk;
-#endif
     int follow_src;		/*  -p[r[+[+]]]		:3	*/
     int follow_href;		/*  -r[+[+]]		:3	*/
     int follows_max;		/*  [-{p,{{pr,r}{,+{,+}}}}]	:3	*/
     int dir_mode;		/*  -nd & -xd		:2	*/
     int max_depth;		/*  -ld			short	*/
     int max_recurse;		/*  -l			short	*/
+    ptrarr_t *filter_list;	/*  -A/-R			*/
+    ptrarr_t *ldom_list;	/*  -D/-Dl			*/
+    ptrarr_t *rdom_list;	/*  -D/-Dr			*/
     off_t max_bytes;		/*  -lb				*/
+    off_t buff_size;		/*  -xs				*/
     int inhibit_cgiget;		/*  -xg & -ng		:2	*/
-    int force_recurse;		/*  -F			:1	*/
+    int force_html;		/*  -F			:1	*/
     int update_mode;		/*  -u, -c & -nc	:2	*/
+    int ext_dump;		/*  -xo & -xO		:2	*/
+    int dump_refs;		/*  -xr			:1	*/
     disp_path_t *disp_path;	/*  -P				*/
     char *index_filename;	/*  -xi				*/
-    int enumerate_files;	/*  -xe			:1	*/
 #ifndef HAVE_CYGWIN
     int fat_quotes;		/*  -xq			:1	*/
 #endif
@@ -208,52 +221,55 @@ typedef struct options_t {
     int send_if_range;		/*  -ni			:1	*/
     int send_referer;		/*  -nR			:1	*/
     int uar_total;		/*  [-U & -iU]			*/
-    ptrarr_t user_agents;	/*  -U & -iU			*/
-    ptrarr_t aux_headers;	/*  -xH				*/
+    ptrarr_t *user_agents;	/*  -U & -iU			*/
+    ptrarr_t *aux_headers;	/*  -xH				*/
+    ptrarr_t *save_headers;	/*  -xh				*/
     int timeout_connect;	/*  -Tc			short	*/
     int timeout_data;		/*  -Td			short	*/
     int max_attempts;		/*  -t			short	*/
     int fail_no_wait;		/*  -nw			:1	*/
-    ptrarr_t bind_addrs;	/*  -xb & ib			*/
-    ptrarr_t proxies;		/*  -y & -iy			*/
+    int http_err_trans;		/*  -xT			:1	*/
+    ptrarr_t *bind_addrs;	/*  -xb & ib			*/
+    proxyarr_t *proxies;	/*  -y & -iy			*/
+
+    char *url_prefix;		/*  -B				*/
+    int enumerate_urls;		/*  -xE			:1	*/
+
+    u_int cow:1;
 } options_t;
 
 
-/*  url spec.  */
+typedef struct disp_t {
+    u_char multi:1;		/*  -r, -p or multiple URLs	*/
+    u_char created:1;
+    u_char devnull:1;
+    char disp[0];
+} disp_t;
 
+/*  shared url parameters  */
 typedef struct url_parm_t {
-#ifdef USE_MAGIC
-    int chk;
-#endif
     options_t *opt;
     char *http_auth;		/*  User:Password in BASE64 encoding  */
-    char *disposition;		/*  -O, [-xE] & URL*disp	*/
-    proxy_t *proxy;		/*  URL^proxy			*/
-    uint ref_count:31;		/*  >1 -> actually copy in clone_parm() */
-    uint strictproxy:1;		/*  URL^^proxy			*/
+    disp_t *disposition;	/*  -O, [-xE]			*/
+    proxy_t *proxy;		/*  -xy, -xyy			*/
+    time_t time_stamp;		/*  -xu				*/
+    uint ref_count:31;		/*  >1 -> actually copy in detach_parm() */
+    uint strictproxy:1;		/*  -xyy			*/
 } url_parm_t;
 
 typedef struct url_t {
-#ifdef USE_MAGIC
-    u_int attempt;		/*  Nr of attempt to download this URL  */
-    int chk;			/*  Structure checksum  */
-    /* checked data begins here  */
-    u_int len;			/*  Structure lenght  */
-#endif
     struct url_t *next;		/*  next url in global chain  */
     struct url_t *referer;	/*  Referer URL (or NULL)  */
 
     url_parm_t *parm;		/*  URL parameters  */
     u_int
-#ifndef USE_MAGIC
 	attempt:8,		/*  Nr of attempt to download this URL  */
-#endif
 	is_top_dir:1,		/*  Get disp_off from this url?  */
 	is_requisite:1,		/*  1=comes from src=, otherwise href=  */
+	save_content:2,		/*  May be unset by -A/-R */
 	relocs:3;		/*  How often this URL was redirected  */
 
     /*  Url: host, port and local part:  */
-    int url_hash;		/*  (For quick search)  */
     host_t *host;		/*  Host of URL  */
     u_short port;		/*  Host port  */
     short link_depth;		/*  For -l */
@@ -262,25 +278,22 @@ typedef struct url_t {
     char local_part[0];		/*  Variable length local part  */
 } url_t;
 
-/*  queue element for waiting urls  */
-
+/*  queue element for waiting urls in download queues  */
 typedef struct wurl_t {
     struct wurl_t *next;	/*  next url in current queue  */
 
     url_t *url;			/*  the associated url  */
 } wurl_t;
 
-/*  queue element for waiting objects  */
 
-typedef struct wobj_t {
-    struct wobj_t *next;	/*  next object in current queue  */
-
-    url_t *url;			/*  the associated url (if any)  */
-    proxy_t *proxy;		/*  the associated proxy (if any)  */
-} wobj_t;
+/*  buffer element  */
+typedef struct buffe_t {
+    struct buffe_t *next;
+    size_t len;
+    char data[0];
+} buffe_t;
 
 /*  queue element for active urls  */
-
 typedef struct aurl_t {
     struct aurl_t *next;	/*  next url in current queue  */
 
@@ -296,6 +309,7 @@ typedef struct aurl_t {
     int socket;			/*  TCP socket  */
     int ipidx;			/*  Index of the used haddr_t  */
 /*    int bipidx;	*/		/*  Index of the bound outgoing ip  */
+    proxyent_t *proxye;		/*  Pointer to used proxy entry  */
     proxy_t *proxy;		/*  Pointer to used proxy  */
     int pipidx;			/*  Index of the proxy's used haddr_t  */
     
@@ -304,54 +318,81 @@ typedef struct aurl_t {
     u_int
 	file_created:1,		/*  Did we already create the file?  */
 	http_done_header:1,	/*  End of header reached?  */
+	content_is_text:1,	/*  Dump to multi-dest?  */
 	content_is_html:1,	/*  Try recursive search on file?  */
 	reloc:1;		/*  Is this url being redirected?  */
+
+    char *auth_chall;		/*  Challenge for auth_lev  */
     
+    char *headers;		/*  HTTP headers for saving to file  */
+    int hdrslen, hdrssiz;	/*  Actual and allocated lenght of above  */
+
     /*  Input data overlap buffer  */
     char *buffer;		/*  Overlap buffer  */
     short size;			/*  Current size of overlap buffer  */
     short offset;		/*  Current fill state of overlap buffer  */
 
+    buffe_t *buff;		/*  Data to writeout; circular_queue  */
+    size_t buff_len;		/*  To prevent chewing up all mem  */
+
     short http_result_code;	/*  200, 400 etc. 0 means no status code by now  */
 
     short displen;		/*  Length of >>  */
     char disposition[0];	/*  Local file name  */
 } aurl_t;
 
+/*  parsed url  */
+typedef struct {
+/*    const char *srct;*/		/*  only for error messages  */
+    host_t *host;
+    const char *auth;
+    const char *lpart;
+    int authlen;
+    int lpartlen;
+    int pathlen;
+    u_short port;
+/* #define PR_HTTP	-1 */
+/*    short proto; */		/*  >= 0 => strlen(proto_string)  */
+} purl_t;
+
+/*  queue element for hosts waiting in lookup queues  */
 typedef struct whost_t {
     struct whost_t *next;	/*  fork/read queue link  */
-
-    wobj_t *objq;		/*  urls/proxies depending on this lookup  */
     host_t *host;		/*  host being looked up  */
-    int fd;			/*  pipe to read from  */
-    int pid;			/*  pid of the DNS helper process  */
-    long timeout;		/*  lookup timeout  */
+
+    url_t *urls;		/*  urls depending on this lookup; circular queue  */
+    int num_proxies;		/*  number of proxies depending on this lookup  */
 } whost_t;
 
-/* set: prepend, remove random */
-#define ex_linear_na_queue(q, t) extern t *q
-#define linear_na_queue(q, t) t *q
-#define lnq_append(q, e) do {e->next = q; q = e;} while(0)
-/* #define lnq_rm1st(q) do {q = q->next;} while(0) */
-#define lnq_remove(q, e) do {*e##p = (*e##p)->next;} while(0)
-#define lnq_iterate(q, t, e, a) do {t *e; for (e = q; e; e = e->next) a;} while(0)
-#define lnq_iterate_rm(q, t, e, a, c) do {t *e, **e##p; for (e##p = &q; (e = *e##p); ) {a; c: e##p = &(e->next); }} while(0)
+typedef struct dnsproc_t {
+    struct dnsproc_t *next;
 
-/* queue: append, remove first */
-#define ex_linear_queue(q, t) extern t *q, **q##_app
-#define linear_queue(q, t) t *q, **q##_app = &q
-#define lq_append(q, e) do {e->next = 0; *q##_app = e; q##_app = &(e->next);} while(0)
-#define lq_rm1st(q) do {q = q->next; if (!q) q##_app = &q; } while(0)
-/* #define lq_remove(q, e) do {if (q##_app == &((*e##p)->next)) q##_app = e##p; *e##p = (*e##p)->next;} while(0) */
-#define lq_consume(q, t, e, a) do {t *e; while (q) {e = q; a;}} while(0)
+    whost_t *whost;		/*  host being looked up  */
+    int fd;			/*  control socket  */
+    int pid;			/*  pid of the DNS helper process  */
+    long timeout;		/*  lookup or idle timeout  */
+} dnsproc_t;
+
+
+/*
+ * notes on the iterators:
+ * - removal must be done before the set/queue is possibly extended
+ * - new elements will be missed by the iterator
+ */
+
+/* set: prepend, remove random */
+#define ls_add(q, e) do {e->next = q; q = e;} while(0)
+#define ls_iterate(q, t, e, a) do {t *e; for (e = q; e; e = e->next) a} while(0)
+#define ls_iterate_rm(q, t, e, a) do {t *e, **e##p; for (e##p = &q; (e = *e##p); ) {a e##p = &e->next; }} while(0)
+#define ls_remove(e) *e##p = e->next
 
 /* circular queue: append, remove first, move first to end */
-#define ex_circular_queue(q, t) extern t *q
-#define circular_queue(q, t) t *q
-#define cq_append(q, e) do {if(q) {e->next = q->next; q->next = e;} else e->next=e; q = e;} while(0)
+#define cq_append(q, e) do {if(q) {e->next = q->next; q->next = e;} else e->next = e; q = e;} while(0)
+#define cq_prepend(q, e) do {if(q) {e->next = q->next; q->next = e;} else {e->next = e; q = e;}} while(0)
+#define cq_iterate(q, t, e, a) do {if (q) {t *e = q; do {e = e->next; a} while (e != q);}} while(0)
+#define cq_consume(q, t, e, a) do {t *e, *e##fp; if (q) {e##fp = q; do {e = q->next; a} while (e != e##fp);}} while(0)
 #define cq_rm1st(q) do {if(q->next == q) q = 0; else q->next = q->next->next;} while(0)
-/* #define cq_remove(q, e) do {if(q->next == q) q = 0; else {if(e##p->next == q) q = e##p; e##p->next = e##p->next->next;}} while(0) */
-#define cq_consume(q, t, e, a) do {t *e, *e##fp; if (q) {e##fp = q; do {e = q->next; a;} while (e != e##fp);}} while(0)
+
 
 /*  Update modes for already existing files  */
 #define EX_CLOBBER	0
@@ -365,8 +406,8 @@ typedef struct whost_t {
 #define DIRS_ALWAYS	2
 
 /*  Levels of url recursion:  */
-#define DONT_FETCH		0
-#define NOT_RECURSIVE		1
+#define NOT_RECURSIVE		0
+#define SAMEDIR_RECURSIVE	1
 #define SUBDIR_RECURSIVE	2
 #define HOST_RECURSIVE		3
 #define GLOBAL_RECURSIVE	4
@@ -380,23 +421,22 @@ typedef struct whost_t {
 #define RT_DONE		5	/* finished */
 #define RT_TIMEOUT	6	/* proxy->server timeout */
 #define RT_REFUSED	7	/* proxy->server refusal */
+#define RT_RESTART	8	/* immediate retry */
 
 /*  Functions:  */
 
 /*  main.c:  */
 extern char *progname;
 extern int verbose;
-int calc_hash(u_char * data, int len);
-#ifdef __DMALLOC_H__
-# define mmalloc malloc
-# define mrealloc realloc
-#else
+int calc_nhash(const u_char *data, int len);
+int calc_hash(const u_char *data);
 void *mmalloc(size_t size);
 void *mrealloc(void *ptr, size_t size);
-#endif
-void die(int ret, char *msg, ...);
-int errm(url_t *u, char *txt, ...);
-void prx(int lev, char *txt, ...);
+void die(int ret, const char *msg, ...);
+int errm(url_t *u, const char *txt, ...);
+int verrm(url_t *u, const char *txt, va_list va);
+void prx(int lev, const char *txt, ...);
+void prxu(int lev, url_t *u, const char *txt, ...);
 
 #define NFO 3
 #define WRN 2
@@ -404,13 +444,19 @@ void prx(int lev, char *txt, ...);
 
 #ifdef DEBUG
 extern int debug;
-void dbp(char *txt, ...);
-void dbpe(char *txt, ...);
+void dbp(const char *txt, ...);
+void dbpe(const char *txt, ...);
+void dbpu(url_t *u, const char *txt, ...);
+void dbpeu(url_t *u, const char *txt, ...);
 # define dbg(wht, tdo) do { if (wht & debug) dbp tdo; } while (0)
 # define dbge(wht, tdo) do { if (wht & debug) dbpe tdo; } while (0)
+# define dbgu(wht, tdo) do { if (wht & debug) dbpu tdo; } while (0)
+# define dbgeu(wht, tdo) do { if (wht & debug) dbpeu tdo; } while (0)
 #else
 # define dbg(wht, tdo)
 # define dbge(wht, tdo)
+# define dbgu(wht, tdo)
+# define dbgeu(wht, tdo)
 #endif
 
 #define URL 1
@@ -418,7 +464,7 @@ void dbpe(char *txt, ...);
 #define QUE 4
 #define CON 8
 #define HDR 16
-#define CHK 32
+#define REF 32
 #define MEM 64
 
 #if 0
@@ -427,97 +473,171 @@ void dbpe(char *txt, ...);
 # define ierr(m) *(char *)0 = 0
 #endif
 
-#ifdef USE_MAGIC
-void magck(const char *);
-# define checken(m) magck(m)
-# define ichecken_hashurl(p) calc_hash((char *)&(p->len), p->len)
-# define ichecken_hashgen(p) calc_hash((char *)p+sizeof(int), sizeof(*p)-sizeof(int))
-# define checken_updurl(p) p->chk = ichecken_hashurl(p)
-# define checken_updgen(p) p->chk = ichecken_hashgen(p)
-# define ichecken_url(p) (p->chk == ichecken_hashurl(p))
-# define ichecken_gen(p) (p->chk == ichecken_hashgen(p))
-# define checken_url(p,m) do { if(!ichecken_url(p)) checken(m); } while(0)
-# define checken_gen(p,m) do { if(!ichecken_gen(p)) checken(m); } while(0)
-#else
-# define checken(m) do ; while(0)
-# define checken_updurl(p) do ; while(0)
-# define checken_updgen(p) do ; while(0)
-# define checken_url(p,m) do ; while(0)
-# define checken_gen(p,m) do ; while(0)
-#endif
+int patternMatch(const char *string, int slen, const char *pattern);
 
 /*  hostlist.c:  */
 extern int always_primary_name;
-extern host_t *hostlist;
-ex_linear_queue(queue_dns_forks, whost_t);
-ex_linear_na_queue(queue_dns_reads, whost_t);
-host_t *host_lookup_fast(char *name, int namlen);
-host_t *host_lookup_full(char *name, int namlen, url_t *u, proxy_t *prx);
-int fork_lookup(whost_t *wh);
-int finish_lookup(whost_t *wh);
+extern whost_t *queue_dns_lookup;
+extern dnsproc_t *list_dns_busy, *list_dns_idle;
+host_t *host_lookup_fast(const char *name, int namlen);
+whost_t *host_lookup_full(const char *name, int namlen);
+int start_lookup(dnsproc_t *pr);
+int finish_lookup(dnsproc_t *pr);
+dnsproc_t *fork_dnsproc(void);
+void reap_dnsproc(dnsproc_t *pr);
 
 /*  getopts.c:  */
 void getopts(int argc, char *argv[]);
 
 /*  fetch.c:  */
 extern off_t max_bytes, fetched_bytes, total_bytes;
-extern int max_dns_forks;
+extern int max_dnss_active;
 extern int max_urls_active;
 extern int timeout_dns;
 extern int max_time;
 extern int max_urls;
 extern int num_urls;
+extern int num_urls_active;
 extern int num_urls_done;
 extern int num_urls_fail;
 extern int num_errors;
 extern int show_stat;
 extern int waiting_proxies;
 extern int all_proxy_wait;	/* unused */
-extern struct timeval cur_tv;
+extern struct timeval cur_tv, throttle;
 extern struct sockaddr_in bind_addr;
-ex_circular_queue(queue_urls_connect, wurl_t);
-ex_linear_na_queue(queue_urls_request, aurl_t);
-ex_linear_na_queue(queue_urls_reply, aurl_t);
+extern wurl_t *queue_urls_connect;
+extern aurl_t *list_urls_request, *list_urls_reply;
 int touch(aurl_t *au);
-void cleanup(char *msg);
+void byebye(const char *msg);
 void fetch_all(void);
 
 /*  url.c:  */
 extern int economize_dns;
-extern url_t *urllist;
-extern proxy_t *proxylist;
-int same_dir(char *path, int len, url_t *referer, int is_req);
-int find_url(char *path, int len, hinfo_t *hinfo, u_short port, int *hashp);
-proxy_t *parse_proxy (char *proxy, int ratio);
-int parse_add_url(char *url, int len, url_t *referer, url_parm_t *parm,
-                  int istopdir, int isreq, int relocs, int link_depth);
+int same_host(url_t *referer, host_t *host, u_short port, int is_req);
+int same_dir(const char *path, int len, url_t *referer, int is_req);
+int find_url(const char *path, int len, hinfo_t *hinfo, u_short port, int *hashp);
+proxy_t *parse_add_proxy(const char *srct, const char *proxy);
+int test_pat(const char *path, int len, int fp,
+	     const char *mtype, int mtl, url_parm_t *parm);
+int print_url(char *buf, int bufl, url_t *u, int auth);
+int parse_add_url(const char *srct, const char *url, int len, url_t *base,
+		  url_t *referer, url_parm_t *parm,
+		  int isreq, int relocs, int link_depth,
+		  aurl_t *au);
+void finish_whost(whost_t *wh);
 int queue_url(url_t *u);
-void add_url(url_t *u);
+void add_url(url_t *u, int hash);
 void free_url(url_t *u);
-int clone_parm(url_t *u);
+int detach_parm(url_t *u);
 
 /*  recurse.c:  */
-int recurse_buff(url_t *u, char *databuf, int len, int notlast);
-void recurse_pfile(url_t *u, int fi, char **bupo, int *lepo);
+int recurse_buff(url_t *u, const char *databuf, int len, int notlast, aurl_t *au);
+void recurse_pfile(url_t *u, int fi, char **bupo, int *lepo, aurl_t *au);
 void recurse_file(url_t *u, char *name);
+int needs_recurse_u(url_t *u, int dump);
+int needs_recurse_au(aurl_t *au, int dump);
 
 /*  http_conn.c:  */
 int activate_url(url_t *u, aurl_t **au);
+int connect_url(aurl_t *au, struct timeval *next_conn_tv);
+void disconnect_url(aurl_t *au);
+void deactivate_url(aurl_t *au);
 
 /*  http_req.c:  */
-extern void init_user_agent(void);
+void init_user_agent(void);
 #define len_enc_auth(x) (4 * (((x) + 2) / 3) + 1)
-void encode_auth(char *buf, char *auth, int len);
+void encode_auth(char *buf, const u_char *auth, int len);
+#define len_dec_auth(x) (3 * (((x) + 3) / 4) + 1)
+int decode_auth(char *p, const u_char *s);
 int send_http_get(aurl_t *au);
 
 /*  http_rsp.c:  */
 extern int economize_files;
+extern size_t max_buffer;
 int mmfopen(char *name, int flags, int *f);
-int free_fd(void);
+int free_fd(int steallast);
+int werrm(aurl_t *au, int sts, const char *msg, ...);
+int uwerrm(url_t *u, int sts, const char *msg, ...);
+void write_usts(url_t *u, int sts);
+void write_psts(url_parm_t *parm, const char *u, int ul, int istopdir, int sts);
 int handle_reply(aurl_t *au);
 
 /*  util_date.c:  */
 #define BAD_DATE 0
 time_t parseHTTPdate(const char *date);
 
+
+#define cat_chr(b,l,c) b[l++] = c
+#define cat_mem(b,l,ml,m) do { int _ml = ml; memcpy(b + l, m, _ml); l += _ml; } while(0)
+#define cat_str(b,l,s) do { const char *_s = s; cat_mem(b, l, strlen(_s), _s); } while(0)
+
+#define lcat_chr(b,bl,l,c) do { if (l < bl) b[l] = c; l++; } while(0)
+#define lcat_mem(b,bl,l,ml,m) do { int _ml = ml; if (l + _ml < bl) memcpy(b + l, m, _ml); l += _ml; } while(0)
+#define lcat_str(b,bl,l,s) do { const char *_s = s; lcat_mem(b, bl, l, strlen(_s), _s); } while(0)
+
+
+#define BUCKET_SHIFT 6
+
+static inline int
+hash_shift(int num)
+{
+    int i;
+
+#if defined(__GNUC__) && defined(__i386__)
+    asm("xorl %0,%0\n\tbsr %1,%0" : "=r" (i) : "rm" (num));
+    i -= BUCKET_SHIFT - 1;
+    if (i < 0)
+	i = 0;
+#else
+    for (i = 0; num >= (1 << BUCKET_SHIFT << i); i++);
+#endif
+    return i;
+}
+
+
+#define h_find(str, strc, strt, hash, fcmp, ffound) \
+    do { \
+	strt *s; \
+	int osz = hash_shift(strc); \
+	for (s = osz ? ((strt **)str)[hash & ((1 << osz) - 1)] : str; \
+	     s; s = s->next) \
+	{ \
+	    if (fcmp(s)) { \
+		ffound(s) \
+	    } \
+	} \
+    } while(0)
+
+#define h_add(str, strc, strt, s, hash, ferr, fhash) \
+    do { \
+	strt *sp, *nsp, **spp, **nspp; \
+	int i, osz, nsz, nhash; \
+ \
+	osz = hash_shift(strc); \
+	strc++; \
+	nsz = hash_shift(strc); \
+	if (osz != nsz) { \
+	    if (!(nspp = mmalloc((1 << nsz) * sizeof(strt *)))) { \
+		ferr \
+	    } else { \
+		memset(nspp, 0, (1 << nsz) * sizeof(strt *)); \
+		for (i = 0; i < (1 << osz); i++) \
+		    for (sp = osz ? ((strt **)str)[i] : str; sp; sp = nsp) { \
+			nsp = sp->next; \
+			nhash = fhash(sp); \
+			spp = nspp + (nhash & ((1 << nsz) - 1)); \
+			sp->next = *spp; \
+			*spp = sp; \
+		    } \
+		if (osz) \
+		    free(str); \
+		str = (strt *)nspp; \
+	    } \
+	} \
+	spp = nsz ? ((strt **)str) + (hash & ((1 << nsz) - 1)) : &str; \
+	s->next = *spp; \
+	*spp = s; \
+    } while(0)
+
 #endif				/*  _PUF_H  */
